// better selection between compressed and unpropressed files  

// Program made by DJ7OO ( dj7oo@t-online.de ) with using
// LoRa library and program sample by Stuart Robinson, GW7HPW
// https://github.com/StuartsProjects/SX12XX-LoRa

#include <SPI.h>           //the SX127X device is SPI based so load the SPI library
#include <SX127XLT.h>      //include the appropriate library   
#include "Settings.h"      //include the settings file, frequencies, LoRa settings etc   
#include "BluetoothSerial.h"
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <Arduino.h>

// DISPLAY address
#define SSD1306_ADDRESS 0x3C
#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels
#define OLED_RESET 16 

Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

BluetoothSerial bluetooth;

SX127XLT LT;   //create a library class instance called LT

uint32_t RXpacketCount;
uint32_t errors;
uint8_t RXBUFFER[RXBUFFER_SIZE]; //create the buffer that received packets are copied into
uint8_t RXPacketL;               //stores length of packet received
int8_t  PacketRSSI;              //stores RSSI of received packet
int8_t  PacketSNR;               //stores signal to noise ratio of received packet
char Packet_p;
bool APRS_Mode;

String Tcall, Instring, Outstring;
String xLat = "";
String Ns2 = "";
String xLon = "";
String Ew2 = "";
String APRS = "";
char tab;
char symb;
int cnt = 1;
String Alt, Ns, Ew, addtext, aprsdata;
float Tlat, Tlon, Talt, Lat, Lon;

/////////////////////////////////////////////////////////////////////////////
void loop()
{
  RXPacketL = LT.receive(RXBUFFER, RXBUFFER_SIZE, 60000, WAIT_RX); //wait for a packet to arrive with 60seconds (60000mS) timeout
  if (BUZZER > 0)                //turn buzzer on
  {
  digitalWrite(BUZZER, HIGH);
  }
  PacketRSSI = LT.readPacketRSSI(); //read the recived RSSI value
  PacketSNR = LT.readPacketSNR();   //read the received SNR value
  if (RXPacketL == 0)               //if the LT.receive() function detects an error, RXpacketL == 0
  {
    packet_is_Error();
  }
  else
  {
    packet_is_OK();
  }
  if (BUZZER > 0)
  {
    digitalWrite(BUZZER, LOW);      //buzzer off
  }
  Serial.println();
}

/////////////////////////////////////////////////////////////////////////
void packet_is_OK()
{
  digitalWrite(LED1, HIGH);      //something has happened
  uint16_t IRQStatus, localCRC;
  IRQStatus = LT.readIrqStatus();   //read the LoRa device IRQ status register
  RXpacketCount++;
 
// making a string out of incoming LoRa data 
  byte lora_LLoop;
  byte lora_LData;
  byte lcount;
  for (lora_LLoop = 0; lora_LLoop <= RXPacketL; lora_LLoop++)
  {
  lcount++;
  lora_LData = RXBUFFER[lora_LLoop];
  Instring += char(lora_LData);     
  }

  Instring =Instring.substring(3);
  Serial.println(Instring);
  int posx = Instring.indexOf(">"); 
  Tcall = Instring.substring(0,posx);
  Serial.print("Call: ");
  Serial.println(Tcall); 

  cnt=cnt+1;
  if(cnt>999) { cnt=1; } 

  decode_aprs();
 
  display.clearDisplay();
  display.setTextColor(WHITE);
  display.setTextSize(2);
  int posc1 = Tcall.length();     // calc for 
  int posc2 = (128-(posc1*12))/2; // centering
  display.setCursor(posc2,0);
  display.println(Tcall);
  display.fillRect(0,20,160,2,WHITE);
  display.setCursor(0,30);
  display.setTextSize(1);
  display.print("RSSI: ");
  display.setTextSize(2);
  display.print(PacketRSSI);
  display.println("dbm");
  display.setCursor(0,50);
  display.setTextSize(1);
  display.print("SNR : ");
  display.setTextSize(2);
  display.print(PacketSNR);
  display.println("db");
  display.setTextSize(1);
  display.setCursor(110,56);
  if(cnt<100) { display.print("0"); }
  if(cnt<10)  { display.print("0"); }
  display.print(cnt);  
  display.display();  
    
  localCRC = LT.CRCCCITT(RXBUFFER, RXPacketL, 0xFFFF);  //calculate the CRC, this is the external CRC calculation of the RXBUFFER
//  Serial.println(F(" "));
  Serial.print(F("RSSI: "));
  Serial.print(PacketRSSI);
  Serial.println(F("dBm"));
  Serial.print(F("SNR : "));
  Serial.print(PacketSNR);
  Serial.println(F("dB"));

  Instring = "";
  digitalWrite(LED1, LOW);  
}

/////////////////////////////////////////////////////////////////////////////
void packet_is_Error()
{
  uint16_t IRQStatus;
  IRQStatus = LT.readIrqStatus();   //read the LoRa device IRQ status register
  if (IRQStatus & IRQ_RX_TIMEOUT)   //check for an RX timeout
  {
  }
  else
  {
    errors++;
    Serial.print(F(" PacketError"));
    Serial.print(F(",RSSI,"));
    Serial.print(PacketRSSI);
    Serial.print(F("dBm,SNR,"));
    Serial.print(PacketSNR);
    Serial.print(F("dB,Length,"));
    Serial.print(LT.readRXPacketL());  //get the real packet length
    Serial.print(F(",Packets,"));
    Serial.print(RXpacketCount);
    Serial.print(F(",Errors,"));
    Serial.print(errors);
    Serial.print(F(",IRQreg,"));
    Serial.print(IRQStatus, HEX);
    LT.printIrqStatus();               //print the names of the IRQ registers set
  }
  delay(250);                          //gives a longer buzzer and LED flash for error
}

///////////////////////////////////////////////////////////////////////////////
void printElapsedTime()
{
  float seconds;
  seconds = millis() / 1000;
  Serial.print(seconds, 0);
  Serial.print(F("s"));
}

//////////////////////////////////////////////////////////////////////////////
void led_Flash(uint16_t flashes, uint16_t delaymS)
{
  uint16_t index;
  for (index = 1; index <= flashes; index++)
  {
    digitalWrite(LED1, HIGH);
    delay(delaymS);
    digitalWrite(LED1, LOW);
    delay(delaymS);
  }
}

//////////////////////////////////////////////////////////////////////////////
void setup()
{
  bluetooth.begin("TTGO_RX_3"); 
  delay(3000);
  pinMode(LED1, OUTPUT);   //setup pin as output for indicator LED
  led_Flash(2, 125);       //two quick LED flashes to indicate program start
  Serial.begin(9600);
  Serial.println();
  Serial.print(F(__TIME__));
  Serial.print(F(" "));
  Serial.println(F(__DATE__));
  Serial.println();
  Serial.println(F("LoRa_Receiver_ESP32 Starting"));
  Serial.println();
  
//  Wire.begin(21,22); 
  display.begin(SSD1306_SWITCHCAPVCC, 0x3c); 
  display.clearDisplay();
  display.setTextColor(WHITE);
  display.drawRect(0,0,128,62,WHITE);
  display.setCursor(35,5);
  display.setTextSize(2);
  display.println("LoRa-");
  display.setCursor(20,24);
  display.println("DECODER");
  display.setCursor(5,43);
  display.println("433.775MHz");
  display.display();  
  
  if (BUZZER > 0)
  {
    pinMode(BUZZER, OUTPUT);
    digitalWrite(BUZZER, HIGH);
    delay(50);
    digitalWrite(BUZZER, LOW);
  }

//SPI.begin();                     //default setup can be used be used
  SPI.begin(SCK, MISO, MOSI, NSS); //alternative format for SPI3, VSPI; SPI.begin(SCK,MISO,MOSI,SS)
  if (LT.begin(NSS, NRESET, DIO0, DIO1, DIO2, LORA_DEVICE))
  {
    Serial.println(F("LoRa Device found"));
    led_Flash(2, 125);
    delay(1000);
  }
  else
  {
    Serial.println(F("No device responding"));
    while (1)
    {
      led_Flash(50, 50);                                       //long fast speed LED flash indicates device error
    }
  }

  //***************************************************************************************************
  //Setup LoRa device
  //***************************************************************************************************
  LT.setMode(MODE_STDBY_RC);                              //got to standby mode to configure device
  LT.setPacketType(PACKET_TYPE_LORA);                     //set for LoRa transmissions
  LT.setRfFrequency(Frequency, Offset);                   //set the operating frequency
  LT.calibrateImage(0);                                   //run calibration after setting frequency
  LT.setModulationParams(SpreadingFactor, Bandwidth, CodeRate, LDRO_AUTO);  //set LoRa modem parameters
  LT.setBufferBaseAddress(0x00, 0x00);                    //where in the SX buffer packets start, TX and RX
  LT.setPacketParams(8, LORA_PACKET_VARIABLE_LENGTH, 255, LORA_CRC_ON, LORA_IQ_NORMAL);  //set packet parameters
  LT.setSyncWord(LORA_MAC_PRIVATE_SYNCWORD);              //syncword, LORA_MAC_PRIVATE_SYNCWORD = 0x12, or LORA_MAC_PUBLIC_SYNCWORD = 0x34
  LT.setHighSensitivity();                                //set for highest sensitivity at expense of slightly higher LNA current
  LT.setDioIrqParams(IRQ_RADIO_ALL, IRQ_RX_DONE, 0, 0);   //set for IRQ on RX done
  //***************************************************************************************************

  Serial.println();
  LT.printModemSettings();                                     //reads and prints the configured LoRa settings, useful check
  Serial.println();
  LT.printOperatingSettings();                                 //reads and prints the configured operting settings, useful check
  Serial.println();
  Serial.println();
  Serial.print(F("Receiver ready"));
  Serial.println();
}

//////////////////////////////////////
void decode_aprs()
{
//    ********   select type of protocol   **********    
     int cnt2 = Instring.indexOf("!/"); //compressed
     int cnt3 = Instring.indexOf(":!"); //uncompressed 
     int cnt4 = Instring.indexOf(":="); //uncompressed

//     Serial.print("CNT2: ");Serial.println(cnt2);
//     Serial.print("CNT3: ");Serial.println(cnt3);
//     Serial.print("CNT4: ");Serial.println(cnt4);
            
     if(cnt2!=-1&&(cnt3!=-1||cnt4!=-1)) 
     {

// *********** decoding compressed ( BASE91 ) ***********
      aprsdata = Instring.substring(cnt2+1,cnt2+14);
      addtext = Instring.substring(cnt2+14);    

//   Serial.println(aprsdata);
//   Serial.println(addtext);
     
          tab=aprsdata.charAt(0);
     int lat1=aprsdata.charAt(1);
     int lat2=aprsdata.charAt(2);
     int lat3=aprsdata.charAt(3);
     int lat4=aprsdata.charAt(4);
     int lon1=aprsdata.charAt(5);
     int lon2=aprsdata.charAt(6);
     int lon3=aprsdata.charAt(7);
     int lon4=aprsdata.charAt(8);
         symb=aprsdata.charAt(9);
     int alt1=aprsdata.charAt(10);
     int alt2=aprsdata.charAt(11);
     int alt3=aprsdata.charAt(12);
     
     long Lat1 = (lat1-33)*753571; 
     long Lat2 = ((lat2-33)*82810)/10;
     long Lat3 = (lat3-33)*91;
     long Lat4 = lat4-33;
     float Lat5 = Lat1+Lat2+Lat3+Lat4;
     Tlat = 90-(Lat5/380926);

     if(Tlat<0) { Ns = "S"; } else { Ns = "N"; } 

     if(Tlat < 0) { Tlat= -Tlat; }
     unsigned int Deg_Lat = Tlat; 
     Lat = 100*(Deg_Lat) + (Tlat - Deg_Lat)*60; 
 
     Serial.println("MODE: BASE91");
     Serial.print("Lat:  ");
     Serial.print(Tlat,5);
     Serial.println(Ns);
     
     long Lon1 = (lon1-33)*753571; 
     long Lon2 = ((lon2-33)*82810)/10;
     long Lon3 = (lon3-33)*91;
     long Lon4 = lon4-33;
     float Lon5 = Lon1+Lon2+Lon3+Lon4;
     Tlon = (Lon5/190463)-180;

     if(Tlon<0) { Ew = "W"; } else { Ew = "E"; } 

     if(Tlon < 0) { Tlon= -Tlon; }
     unsigned int Deg_Lon = Tlon; 
     Lon = 100*(Deg_Lon) + (Tlon - Deg_Lon)*60; 

     Serial.print("Lon:  ");
     Serial.print(Tlon,5);
     Serial.println(Ew);
     
     long Alt1 = (alt1-33)*91;            ; 
     long Alt2 = alt2-33;            ;
     float Alt3 = (Alt1+Alt2);
     Alt3=Alt3*0.00086772215;
     Talt = pow(10,Alt3);
     Talt=Talt/3.281;

     Serial.print("Alt:  ");
     Serial.print(Talt,0);
     Serial.println("m");

     Serial.print("Text : ");
     Serial.println(addtext);
     Serial.println(" ");
    
     Instring = "";
     APRS_Mode = 1;
     } 
 
 // *************** decoding uncompressed ****************
     else if(cnt2==-1&&(cnt3!=-1||cnt4!=-1)) 
     { 
         if(cnt3!=-1) { APRS = Instring.substring(cnt3+2); }
    else if(cnt4!=-1) { APRS = Instring.substring(cnt4+2); }

//    Serial.print("APRS: "); Serial.println(APRS);

    xLat = APRS.substring(0,7);
    Lat  = xLat.toFloat();

   Serial.println("MODE: APRS");
   Serial.print("Lat: ");
   Serial.print(Lat,4);

// APRS LAT-Auswertung  
int xlat11 = xLat.charAt(0);
int xlat12 = xLat.charAt(1);
int xlat13 = xLat.charAt(2);
int xlat14 = xLat.charAt(3);
int xlat15 = xLat.charAt(4);
int xlat16 = xLat.charAt(5);
int xlat17 = xLat.charAt(6);

//Serial.print(lat11-48);
//Serial.print(lat12-48);
//Serial.print(lat13-48);
//Serial.print(lat14-48);
//Serial.print(lat16-48);
//Serial.println(lat17-48);

  int lat_grad2 = (xlat11-48)*10+(xlat12-48);
  float lat_min2 = ((xlat13-48)*1000)+((xlat14-48)*100)+((xlat16-48)*10)+(xlat17-48);

    lat_min2 = lat_min2/6000;
//  Serial.println(lat_grad2+lat_min2,4);
    Tlat =lat_grad2+lat_min2,4;
    
    Ns2 = APRS.substring(7,8);

    Serial.println(Ns2);

    xLon = APRS.substring(9,17);
    Lon  = xLon.toFloat();

    Serial.print("Lon: ");
    Serial.print(Lon,4);

// APRS LON-Auswertung  
int xlon11 = xLon.charAt(0);
int xlon12 = xLon.charAt(1);
int xlon13 = xLon.charAt(2);
int xlon14 = xLon.charAt(3);
int xlon15 = xLon.charAt(4);
int xlon17 = xLon.charAt(6);
int xlon18 = xLon.charAt(7);

//Serial.print(lon11-48);
//Serial.print(lon12-48);
//Serial.print(lon13-48);
//Serial.print(lon14-48);
//Serial.print(lon15-48);
//Serial.print(lon17-48);
//Serial.println(lon18-48);

    int lon_grad2 = ((xlon11-48)*100)+(xlon12-48)*10+(xlon13-48);
    float lon_min2 = ((xlon14-48)*1000)+((xlon15-48)*100)+((xlon17-48)*10)+(xlon18-48);
    lon_min2 = lon_min2/6000;
//  Serial.println(lon_grad2+lon_min2,4);
    Tlon = lon_grad2+lon_min2,4;
    
    Ew2 = APRS.substring(17,18);
    Serial.println(Ew2);

    tab  = APRS.charAt(8);
    symb = APRS.charAt(18);
    addtext = APRS.substring(19);
    
//  int pos4a = APRS.indexOf("/");
//  String crs = APRS.substring(pos4a-3,pos3);     // CRS   
//  String spd = APRS_1.substring(pos4a+1,pos3+4); // SPD

    int pos5 = APRS.indexOf("/A=");
    if(pos5!=-1) Alt = APRS.substring(pos5+3,pos5+9);
//  Serial.println(Alt);
    Talt = Alt.toInt(); 
    Talt=Talt/3.281;

    Instring = "";
    APRS = "";
    APRS_Mode = 0;

     if(Tlat<0) { Ns = "S"; } else { Ns = "N"; } 
     if(Tlon<0) { Ew = "W"; } else { Ew = "E"; } 
     
     if(Tlat < 0) { Tlat= -Tlat; }
     unsigned int Deg_Lat = Tlat; 
     Lat = 100*(Deg_Lat) + (Tlat - Deg_Lat)*60; 

     if(Tlon < 0) { Tlon= -Tlon; }
     unsigned int Deg_Lon = Tlon; 
     Lon = 100*(Deg_Lon) + (Tlon - Deg_Lon)*60; 

     Serial.print("Alt: ");
     Serial.print(Talt,0);
     Serial.println("m");

     Serial.print("Text : ");
     Serial.println(addtext);
     Serial.println(" ");
   }

// make a string out of it and send it via Bluetooth
   make_aprs();  //****************************
}

///////////////////////////////////////////////
void make_aprs()
  {
// building APRS Standard sentence
 Outstring = (Tcall);
 Outstring += ">APRS:!";
 if(Tlat<10) {Outstring += "0"; }
 Outstring += String(Lat,2);    
 Outstring += Ns;    
 Outstring += tab;  // APRS table
 if(Tlon<100) {Outstring += "0"; }
 if(Tlon<10) {Outstring += "0"; }
 Outstring += String(Lon,2); 
 Outstring += Ew; 
 Outstring += symb;  // APRS symbol
 Outstring += (" ");
 
 //Outstring += " /A=";    
 //if(Talt<10000) { Outstring += "0"; }
 //if(Talt<1000)  { Outstring += "0"; }
 //if(Talt<100)   { Outstring += "0"; }
 //if(Talt<10)    { Outstring += "0"; }
 //Outstring += String(Talt,0);    
 //Outstring += "m ";
 
 Outstring += (addtext);
 Outstring += (" RSSI:"); 
 Outstring += (PacketRSSI);
 Outstring += (" SNR:");
 Outstring += (PacketSNR);
 Outstring += (" *");
   
// sending Standard APRS sentence via Bluetooth 
// Serial.print("APRS-AUSGANGSDATEN: ");
   Serial.println(Outstring);  
   bluetooth.println(Outstring); 
   Talt = 0;
   APRS = "";
   addtext = "";
   Outstring = "";
  } 
